#include <stdio.h>
#include <stdlib.h>

#ifdef TCL_IRSIM
#include <tk.h>
#endif

#include "defs.h"
#include "net.h"
#include "globals.h"
#include "subckt.h"

#ifdef USER_SUBCKT 

typedef struct _Sub
{
    char	*name;			/* name of this sub-circuit */
    ufun	model;			/* function that models sub-circuit */
    ufun	init;			/* Function to call to init */
    int		inst;			/* Number of instantiations */
    struct _Sub	*next;			/* list of commands in bucket */
} SubCircuit;

char pot2char[N_POTS] = {'L', 'x', 'X', 'H'};	/* Original definitions */
char pot2ch[N_POTS] = {'l', 'u', 'x', 'h'};	/* For use by Tcl procedures */

extern	userSubCircuit	subs[];

#ifdef TCL_IRSIM

extern Tcl_Interp *irsiminterp;

/*------------------------------*/
/* Method 1:			*/
/* Make use of Tcl Hash tables	*/
/*------------------------------*/

Tcl_HashTable substbl;

/* First add any C-compiled user subcircuits */

public void init_subs()
{
    userSubCircuit *s;
    SubCircuit *sl;
    Tcl_HashEntry *he;
    int newptr;

    Tcl_InitHashTable(&substbl, TCL_STRING_KEYS);

    for (s = subs; s->name != NULL; s++)
    {
	sl = (SubCircuit *) malloc(sizeof(SubCircuit));
	sl->name = s->name;
	sl->model = s->model;
	sl->init = s->init;
	sl->inst = 0 ;
	sl->next = NULL;

	he = Tcl_CreateHashEntry(&substbl, s->name, &newptr);
	Tcl_SetHashValue(he, (ClientData)sl);
    }
}

/* Instantiate a user subcircuit */

public	ufun subckt_instantiate(sname, initfunP, instP, nameP)
    char *sname;
    ufun *initfunP;
    int  *instP;
    char **nameP;	/* Pointer to stored name of the subcircuit */
{
    int newptr;
    SubCircuit *s;
    Tcl_HashEntry *he;

    he = Tcl_CreateHashEntry(&substbl, sname, &newptr);
    s = (SubCircuit *)Tcl_GetHashValue(he);

    if (s == NULL)
    {
	/* Assume that this subcircuit exists as a Tcl procedure */

	s = (SubCircuit *) malloc(sizeof(SubCircuit));
	s->name = strdup(sname);
	s->model = NULL;
	s->init = NULL;
	s->inst = 0;
	s->next = NULL;

	Tcl_SetHashValue(he, (ClientData)s);
    }

    s->inst++;
    *initfunP = s->init;
    *instP = s->inst;
    *nameP = s->name;
    return (s->model);
}

#else

/*------------------------------*/
/* Method 2:			*/
/* Simple fixed-size hash table */
/*------------------------------*/

#define SUBSTBLSIZE	16

private SubCircuit *substbl[SUBSTBLSIZE];

private int HashSub();

public void init_subs()
{
    register int n;
    register userSubCircuit *s;
    register SubCircuit *sl;

    for (n=0; n < SUBSTBLSIZE; n++)
	substbl[n] = NULL;
    for (s = subs; s->name != NULL; s++) {
	sl = ( SubCircuit *) malloc(sizeof(SubCircuit));
	sl->name = s->name;
	sl->model = s->model;
	sl->init = s->init;
	sl->inst = 0 ;
	n = HashSub(s->name);
	sl->next = substbl[n];
	substbl[n] = sl;
    }
}


private int HashSub( name )
	char  *name;
{
    register int  hashcode = 0;

    do
	hashcode = (hashcode << 1) ^ (*name | 0x20);
    while( *(++name) );
    return( ((hashcode >= 0) ? hashcode : ~hashcode) % SUBSTBLSIZE );
}

public	ufun subckt_instantiate(sname, initfunP, instP, nameP)
char *sname;
ufun *initfunP;
int  *instP;
char **nameP;	/* Pointer to stored name of the subcircuit */
{
    register int n;
    register SubCircuit *s;

    n = HashSub(sname);
    
    for (s = substbl[n]; s != NULL; s = s->next) {
	if (str_eql(s->name, sname) == 0) {
	    *initfunP = s->init;
	    *instP = s->inst ;
	    *nameP = s->name;
	    s->inst++;
	    return(s->model);
	}
    }
    return((ufun)NULL);
}

#endif	/* !TCL_IRSIM */
	

#define INVERT(x)	((x == LOW) ? HIGH : ((x == HIGH) ? LOW : X))

public void subckt_model_C(t)
tptr t;
{
    int i;
    register short nins = t->subptr->nins;
    register short nouts = t->subptr->nouts;
    register nptr *nodes = t->subptr->nodes;
    uptr udata = t->subptr->udata;
#ifdef TCL_IRSIM
    int result;
#endif
    
    char *in, *out = NULL;
    double *delay;

    in  = ( char *) malloc((nins + 1)*sizeof(char));
    out = ( char *) malloc((nouts + 1)*sizeof(char));
    delay = ( double *) malloc(nouts*sizeof(double));

    for (i = 0; i < nins; i++)
	in[i] = nodes[i]->npot;
    for (i = 0; i < nouts; i++)
	out[i] = nodes[i+nins]->npot;

    if (t->subptr->model != NULL)
	(*t->subptr->model)(in, out, delay, udata);

#ifdef TCL_IRSIM
    /* The Tcl version of IRSIM allows subcircuit to be defined as Tcl	*/
    /* procedures.  This procedure must return a list whose first	*/
    /* element is a string of all the output values, followed by time	*/
    /* delays to all of the outputs.					*/

    for (i = 0; i < nins; i++)
	in[i] = pot2ch[in[i]];
    in[i] = '\0';
    for (i = 0; i < nouts; i++)
	out[i] = pot2ch[out[i]];
    out[i] = '\0';

    result = Tcl_VarEval(irsiminterp, t->subptr->sname, " ", in, " ", out, NULL);
    if (result == TCL_OK)
    {
	int listlen;
	double dval;
	char *newoutvals;
	Tcl_Obj *elem, *rlist;

	rlist = Tcl_GetObjResult(irsiminterp);
	result = Tcl_ListObjLength(irsiminterp, rlist, &listlen);
	if (result == TCL_OK && listlen == 1 + nouts)
	{
	    result = Tcl_ListObjIndex(irsiminterp, rlist, 0, &elem);
	    newoutvals = Tcl_GetString(elem);
	    if (strlen(newoutvals) == nouts)
	    {
		for (i = 0; i < nouts; i++)
		{
		    result = Tcl_ListObjIndex(irsiminterp, rlist, i + 1, &elem);
		    result = Tcl_GetDoubleFromObj(irsiminterp, elem, &delay[i]);
		    out[i] = ch2pot(newoutvals[i]);
		}

		/* Clear the interpreter result */
		Tcl_ResetResult(irsiminterp);
	    }
	    else nouts = 0;
	}
	else nouts = 0;
    }
    else nouts = 0;
#endif

    for (i = 0; i < nouts; i++ ) {
	    if ( out[i] != X_X ) {
	    	QueueFVal(nodes[i+nins],INVERT(out[i]),
			     (double) 1.0, delay[i]);
	    	QueueFVal(nodes[i+nins+nouts],INVERT(out[i]), 
			     (double) 1.0, delay[i]);
	    }
	    else {	/* Tri-state output */
	    	QueueFVal(nodes[i+nins], HIGH,
			     (double) 1.0, delay[i]);
	    	QueueFVal(nodes[i+nins+nouts], LOW,
			     (double) 1.0, delay[i]);
	    }
    }

    free(in); 
    free(out); 
    free(delay);
}
#endif
